<?php

namespace Drupal\menu_tree_compare\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Menu\MenuLinkManagerInterface;
use Drupal\Core\Menu\MenuLinkTreeInterface;
use Drupal\Core\Menu\MenuTreeParameters;
use Symfony\Component\DependencyInjection\ContainerInterface;

class MenuTreeCompareController extends ControllerBase {
  /**
   * @var \Drupal\Core\Menu\MenuLinkTreeInterface
   */
  protected $menuLinkTree;

  /**
   * @var MenuLinkManagerInterface
   */
  protected $menuLinkManager;

  public function __construct(MenuLinkTreeInterface $menu_link_tree, MenuLinkManagerInterface $menu_link_manager) {
    $this->menuLinkTree = $menu_link_tree;
    $this->menuLinkManager = $menu_link_manager;
  }

  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('menu.link_tree'),
      $container->get('plugin.manager.menu.link')
    );
  }

  /**
   * Controller entry point.
   *
   * @param string $from
   * @param string $to
   * @return mixed
   */
  public function compare($from, $to) {
    $build['#theme'] = 'menu_tree_compare_compare';

    $parameters = new MenuTreeParameters();

    $from_tree = $this->menuLinkTree->load($from, $parameters);
    $to_tree = $this->menuLinkTree->load($to, $parameters);

    // Get array with icon / matching node ID mappings.
    $mapping = $this->getMapping($from_tree, $to_tree);

    $from_tree_output = $this->menuLinkTree->build($from_tree);
    $to_tree_output = $this->menuLinkTree->build($to_tree);

    $this->enrichMenuTree($from_tree_output['#items'], $mapping['from']);
    $this->enrichMenuTree($to_tree_output['#items'], $mapping['to']);

    $build['#from_tree'] = $from_tree_output;
    $build['#to_tree'] = $to_tree_output;

    $build['#attached']['library'][] = 'menu_tree_compare/menu_tree_compare.compare';
    return $build;
  }

  /**
   * Gets mapping array with icons and source nodes.
   *
   * @param array $from_tree
   *   Left tree to compare from.
   * @param array $to_tree
   *   Right tree to compare to.
   *
   * @return array
   *   The mapping.
   */
  protected function getMapping(array $from_tree, array $to_tree) {
    $mapping = [];
    // Get mapping.
    $mapping['from'] = $this->getMappingSkeleton($from_tree);
    $mapping['to'] = $this->getMappingSkeleton($to_tree);
    // Get mapping icons.
    self::setMappingIcons($mapping, 'from');
    self::setMappingIcons($mapping, 'to');
    return $mapping;
  }

  /**
   * Gets mapping by source ID/Menu link ID.
   *
   * @param array $tree
   *   Tree to process.
   *
   * @return array
   *   Mapping.
   */
  protected function getMappingSkeleton(array $tree) {
    // Get visible items.
    $mapping = [];
    $items = self::getVisibleItems($tree);
    /** @var \Drupal\Core\Menu\MenuLinkTreeElement $data */
    foreach (array_values($items) as $data) {
      $menu_link_id = $data->link->getPluginId();
      if (!self::isCompatiblePluginId($menu_link_id)) {
        continue;
      }
      $parent_link_id = $data->link->getPluginDefinition()['parent'];
      $source_nid = $this->getSourceNid($menu_link_id);
      $parent_source_nid = $this->getSourceNid($parent_link_id);
      // Map by source NID for easier source_nid matching when comparing left
      // tree to right tree.
      $mapping['source_nid'][$source_nid] = [
        'parent_source_nid' => $parent_source_nid,
        'weight' => $data->link->getWeight(),
      ];

      $mapping['menu_link_id'][$menu_link_id] = [
        // HTML ID to use.
        'id' => 'from-' . $menu_link_id . '-' . $source_nid,
        // Source Node ID.
        'source_nid' => $source_nid,
        // Source Node ID of the parent link.
        'parent_source_nid' => $parent_source_nid,
        // Weight of this link.
        'weight' => $data->link->getWeight(),
      ];
    }
    return $mapping;
  }

  /**
   * Sets mapping icons in the mapping array.
   *
   * @param array $mapping
   *   Mapping array.
   * @param string $type
   *   Can be either "from" or "to".
   */
  protected static function setMappingIcons(array &$mapping, $type) {
    $types = ['from' , 'to'];
    $other_type = array_diff($types, [$type]);
    $other_type = reset($other_type);

    foreach ($mapping[$type]['menu_link_id'] as $menu_link_id => $item) {
      $source_match = is_numeric($item['source_nid']) && in_array($item['source_nid'], array_keys($mapping[$other_type]['source_nid']));
      if (!$source_match) {
        // Show X icon if item is not in other tree at all (match_type == none).
        $icon_name = 'not-found';
      }
      else {
        if ($item['parent_source_nid'] != $mapping[$other_type]['source_nid'][$item['source_nid']]['parent_source_nid']) {
          // Show ! icon if parent doesn't match (parent_patch === FALSE).
          $icon_name = 'no-parent-match';
        }
        else {
          if ($item['weight'] != $mapping[$other_type]['source_nid'][$item['source_nid']]['weight']) {
            // Show ! icon if parent matches but weight doesn't.
            $icon_name = 'no-weight-match';
          }
          else {
            // Show check icon.
            $icon_name = 'ok';
          }
        }
      }
      $icon = base_path() . drupal_get_path('module', 'menu_tree_compare') . '/i/' . $icon_name . '.png';
      $mapping[$type]['menu_link_id'][$menu_link_id] += [
        'icon' => $icon,
      ];
    }
  }

  /**
   * @param string $plugin_id
   *
   * @return bool
   *   Whether the plugin ID is supported.
   */
  protected static function isCompatiblePluginId($plugin_id) {
    $valid = ['menu_link_content', 'menu_link_field'];
    return !empty(array_filter($valid, function ($module) use ($plugin_id) {
      return substr($plugin_id, 0, strlen($module)) != $module;
    }));
  }

  /**
   * Returns an output structure for rendering a menu tree.
   *
   * @param array $tree
   *   A  structure representing the tree as returned from MenuLinkTreeInterface::build().
   * @param array $mapping
   *   The mapping of items on the left and on the right.
   *
   * @return array
   *   A render arrays.
   *
   * @see MenuLinkTreeInterface::build()
   */
  protected function enrichMenuTree(array &$tree, array $mapping) {
    foreach ($tree as $key => $element) {
      if (!self::isCompatiblePluginId($key)) {
        continue;
      }
      /** @var \Drupal\Core\Template\Attribute $attributes */
      $attributes = $tree[$key]['attributes'];
      $nid = $this->getSourceNid($key);
      $attributes->setAttribute('id', 'treecompare-' . uniqid() . '-' . ($nid ? $nid : 'none'));
      $attributes->setAttribute('data-jstree', "{\"icon\":\"" . $mapping['menu_link_id'][$key]['icon'] . "\"}");
      if (!empty($tree[$key]['below'])) {
        $this->enrichMenuTree($tree[$key]['below'], $mapping);
      }
    }
  }

  /**
   * Gets the source node ID for a menu link.
   *
   * @param int $menu_link_id
   *   Menu link ID.
   *
   * @return bool|string
   *   'None' or the node ID.
   */
  protected function getSourceNid($menu_link_id) {
    $definition = $this->menuLinkManager->getDefinition($menu_link_id, FALSE);
    if (!empty($definition['route_parameters']['node'])) {
      return $definition['route_parameters']['node'];
    }
    return FALSE;
  }

  /**
   * Finds out which items are visible.
   *
   * @param array $tree
   *   Tree to be rendered.
   *
   * @return array
   *   Items to process.
   */
  protected static function getVisibleItems(array $tree) {
    $items = [];
    /**  @var \Drupal\Core\Menu\MenuLinkTreeElement  $data */
    foreach ($tree as $key => $data) {
      if ($data->link->isEnabled()) {
        $items[] = $data;
      }
      if (!empty($data->subtree)) {
        $items_deeper = self::getVisibleItems($data->subtree);
        foreach ($items_deeper as $item) {
          $items[] = $item;
        }
      }
    }
    return $items;
  }

}
